from random import random
import torch
import numpy as np
import pandas as pd
import random


from pybasilica.utilities import fixedFilter
from pybasilica.utilities import denovoFilter
from pybasilica.utilities import stopRun
from pybasilica.utilities import initialize_params
from pybasilica.run import multi_k_run

'''
from utilities import fixedFilter
from utilities import denovoFilter
from utilities import stopRun
from utilities import initialize_params
from run import multi_k_run
'''




def pyfit(M, groups, input_catalogue, reference_catalogue, k, lr, steps, phi, delta):
    # M --------------------- dataframe
    # groups ---------------- list
    # input_catalogue ------- dataframe
    # reference_catalogue --- dataframe
    # k --------------------- list
    # lr -------------------- float
    # steps ----------------- integer
    # phi ------------------- float
    # delta ----------------- float

    theta = np.sum(M.values, axis=1)

    params = initialize_params(M, groups, input_catalogue, lr, steps)

    counter = 1
    while True:

        #print("iteration:", counter)

        # k ------- dtype: list
        k_inf, A_inf, B_inf = multi_k_run(params, k)
        # k_inf --- dtype: int
        # A_inf --- dtype: torch.Tensor
        # B_inf --- dtype: torch.Tensor

        # A_inf ----- dtype: torch.Tensor 
        # B_input --- dtype: data.frame
        # theta ----- dtype: numpy
        input_catalogue_sub = fixedFilter(A_inf, input_catalogue, theta, phi)
        # B_input_sub ---- dtype: list
        
        # B_inf -------- dtype: torch.Tensor
        # cosmic_df ---- dtype: dataframe
        input_catalogue_new = denovoFilter(B_inf, reference_catalogue, delta)
        # B_input_new --- dtype: list

        if input_catalogue is None:
            input_catalogue_list = []
        else:
            input_catalogue_list = list(input_catalogue.index)

        if stopRun(input_catalogue_sub, input_catalogue_list, input_catalogue_new):
            signatures_inf = []
            for p in range(k_inf):
                p = int(p)
                signatures_inf.append("D"+str(p+1))
            signatures = input_catalogue_list + signatures_inf
            mutation_features = list(M.columns)

            # alpha
            A_inf_np = np.array(A_inf)
            A_inf_df = pd.DataFrame(A_inf_np, columns=signatures)   # dataframe

            # beta
            if B_inf is None:
                #B_inf_denovo_df = pd.DataFrame(columns=mutation_features)
                B_inf_denovo_df = None
            else:
                B_inf_denovo_np = np.array(B_inf)
                B_inf_denovo_df = pd.DataFrame(B_inf_denovo_np, index=signatures_inf, columns=mutation_features)
            
            B_inf_fixed_df = input_catalogue    # dataframe | None

            return A_inf_df, B_inf_fixed_df, B_inf_denovo_df
            # A_inf_df ---------- dtype: dataframe
            # B_inf_fixed_df ---- dtype: dataframe
            # B_inf_denovo_df --- dtype: dataframe

        #print("input_catalogue_list", input_catalogue_list, "\n", "input_catalogue_sub", input_catalogue_sub, "\n", "input_catalogue_new", input_catalogue_new, "\n")

        if len(input_catalogue_sub + input_catalogue_new)==0:
            input_catalogue = None
            params["beta_fixed"] = None
        else:
            input_catalogue = reference_catalogue.loc[input_catalogue_sub + input_catalogue_new]  # dtype: dataframe
            params["beta_fixed"] = torch.tensor(input_catalogue.values).float()
        
        counter += 1




# from matplotlib import interactive
from rich.console import Console
from rich.table import Table
from rich import box
import time
from rich.progress import Progress, BarColumn, TextColumn, TaskProgressColumn, TimeRemainingColumn, SpinnerColumn, RenderableColumn
from rich.live import Live
from rich.table import Table
from sys import maxsize

from pybasilica.svi import PyBasilica

def single_run(kwargs):
    obj = PyBasilica(**kwargs)
    obj._fit()

    return obj


def fit(x, k_list=[0,1,2,3,4,5], lr=0.05, n_steps=500, enumer="sequential", cluster=None, groups=None, beta_fixed=None, compile_model = False, \
        CUDA = False, enforce_sparsity = False, regularizer = "cosine", reg_weight = 1, reg_bic = False, store_parameters=False, verbose=True, 
        stage = "random_noise", regul_compare = None, seed = 10):

    if isinstance(k_list, list):
        if len(k_list) > 0: pass
        else: raise Exception("k_list is an empty list!")
    elif isinstance(k_list, int):
        k_list = [k_list]
    else: raise Exception("invalid k_list datatype")
    
    kwargs = {
            "x":x,
            "lr":lr,
            "n_steps":n_steps,
            "enumer":enumer,
            "cluster":cluster,
            "groups":groups,
            "beta_fixed":beta_fixed,
            "compile_model":compile_model,
            "CUDA":CUDA,
            "enforce_sparsity":enforce_sparsity,
            "regularizer":regularizer,
            "reg_weight":reg_weight,
            "reg_bic":reg_bic,
            "store_parameters":store_parameters,
            "stage":stage,
            "regul_compare":regul_compare,
            "seed":seed
        }

    if verbose:
    # Verbose run

        console = Console()
        if beta_fixed is None:
            betaFixed = "No fixed signatures"
        elif len(list(beta_fixed.index.values)) > 10:
            betaFixed = f'{len(list(beta_fixed.index.values))} signatures, Too many to fit here'
        else:
            betaFixed = ', '.join(list(beta_fixed.index.values))

        table = Table(title="Information", show_header=False, box=box.ASCII, show_lines=False)
        table.add_column("Variable", style="cyan")
        table.add_column("Values", style="magenta")
        table.add_row("No. of samples", str(int(x.shape[0])))
        table.add_row("learning rate", str(lr))
        table.add_row("k denovo list", ', '.join(map(str, k_list)))
        table.add_row("fixed signatures", betaFixed)
        table.add_row("Max inference steps", str(n_steps))
        console.print('\n', table)

        myProgress = Progress(
            TextColumn('{task.description} [bold blue] inference {task.completed}/{task.total} done'), 
            BarColumn(), 
            TaskProgressColumn(), 
            TimeRemainingColumn(), 
            SpinnerColumn(), 
            RenderableColumn())
        
        with myProgress as progress:

            task = progress.add_task("[red]running...", total=len(k_list))

            minBic = maxsize
            secondMinBic = maxsize
            bestRun, secondBest = None, None
            for k in k_list:
                kwargs["k_denovo"] = k

                try:
                    obj = single_run(kwargs)

                    if obj.bic < minBic:
                        minBic = obj.bic
                        bestRun = obj
                    if obj.bic > minBic and obj.bic < secondMinBic:
                        secondMinBic = obj.bic
                        secondBest = obj
                except:
                    raise Exception("Failed to run for k_denovo:{k}!")
                
                progress.console.print(f"Running on k_denovo={k} | BIC={obj.bic}")
                progress.update(task, advance=1)

            if bestRun is not None:
                bestRun._convert_to_dataframe(x, beta_fixed)
            
            if secondBest is not None:
                secondBest._convert_to_dataframe(x, beta_fixed)
            
        from uniplot import plot
        console.print('\n-------------------------------------------------------\n\n[bold red]Best Model:')
        console.print(f"k_denovo: {bestRun.k_denovo}\nBIC: {bestRun.bic}\nStopped at {len(bestRun.losses)}th step\n")
        plot(
            [bestRun.losses, bestRun.likelihoods], 
            title="Loss & Log-Likelihood vs SVI steps", 
            width=40, height=10, color=True, legend_labels=['loss', 'log-likelihood'], interactive=False, 
            x_gridlines=[0,50,100,150,200,250,300,350,400,450,500], 
            y_gridlines=[max(bestRun.losses)/2, min(bestRun.likelihoods)/2])
        console.print('\n')

    else:
    # Non-verbose run

        minBic = maxsize
        secondMinBic = maxsize
        bestRun, secondBest = None, None
        for k in k_list:
            kwargs["k_denovo"] = k

            try:
                obj = single_run(kwargs)

                if obj.bic < minBic:
                    minBic = obj.bic
                    bestRun = obj
                    if k == k_list[0] and len(k_list)>1:
                        secondMinBic = obj.bic
                        secondBest = obj

                if obj.bic > minBic and obj.bic < secondMinBic and len(k_list)>1:
                    secondMinBic = obj.bic
                    secondBest = obj
            except:
                raise Exception("Failed to run for k_denovo:{k}!")

        if bestRun is not None:
            bestRun._convert_to_dataframe(x, beta_fixed)
        
        if secondBest is not None:
            secondBest._convert_to_dataframe(x, beta_fixed)

    return bestRun, secondBest



'''
#import utilities

import torch
import pyro
import pyro.distributions as dist

from pybasilica import svi
from pybasilica import utilities



#------------------------------------------------------------------------------------------------
# run model with single k value
#------------------------------------------------------------------------------------------------
def single_k_run(params):
    #params = {
    #    "M" :               torch.Tensor
    #    "beta_fixed" :      torch.Tensor | None
    #    "k_denovo" :        int
    #    "lr" :              int
    #    "steps_per_iter" :  int
    #}
    #"alpha" :           torch.Tensor    added inside the single_k_run function
    #"beta" :            torch.Tensor    added inside the single_k_run function
    #"alpha_init" :      torch.Tensor    added inside the single_k_run function
    #"beta_init" :       torch.Tensor    added inside the single_k_run function

    # if No. of inferred signatures and input signatures are zero raise error
    #if params["beta_fixed"] is None and params["k_denovo"]==0:
    #    raise Exception("Error: both denovo and fixed signatures are zero")


    #-----------------------------------------------------
    #M = params["M"]
    num_samples = params["M"].size()[0]

    if params["beta_fixed"] is None:
        k_fixed = 0
    else:
        k_fixed = params["beta_fixed"].size()[0]
    
    k_denovo = params["k_denovo"]

    if k_fixed + k_denovo == 0:
        raise Exception("Error: both denovo and fixed signatures are zero")
    #-----------------------------------------------------

    
    #----- variational parameters initialization ----------------------------------------OK
    params["alpha_init"] = dist.Normal(torch.zeros(num_samples, k_denovo + k_fixed), 1).sample()
    if k_denovo > 0:
        params["beta_init"] = dist.Normal(torch.zeros(k_denovo, 96), 1).sample()

    #----- model priors initialization --------------------------------------------------OK
    params["alpha"] = dist.Normal(torch.zeros(num_samples, k_denovo + k_fixed), 1).sample()
    if k_denovo > 0:
        params["beta"] = dist.Normal(torch.zeros(k_denovo, 96), 1).sample()

    svi.inference(params)

    #----- update model priors initialization -------------------------------------------OK
    params["alpha"] = pyro.param("alpha").clone().detach()
    if k_denovo > 0:
        params["beta"] = pyro.param("beta").clone().detach()

    #----- outputs ----------------------------------------------------------------------OK
    alpha_tensor, beta_tensor = utilities.get_alpha_beta(params)  # dtype: torch.Tensor (beta_tensor==0 if k_denovo==0)
    #lh = utilities.log_likelihood(params)           # log-likelihood
    bic = utilities.compute_bic(params)                     # BIC
    #M_R = utilities.Reconstruct_M(params)           # dtype: tensor
    
    return bic, alpha_tensor, beta_tensor


#------------------------------------------------------------------------------------------------
# run model with list of k value
#------------------------------------------------------------------------------------------------
def multi_k_run(params, k_list):
    
    #params = {
    #    "M" :               torch.Tensor
    #    "beta_fixed" :      torch.Tensor
    #    "lr" :              int
    #    "steps_per_iter" :  int
    #}
    #"k_denovo" : int    added inside the multi_k_run function
    

    bic_best = 10000000000
    k_best = -1

    for k_denovo in k_list:
        try:
            params["k_denovo"] = int(k_denovo)
            bic, alpha, beta = single_k_run(params)
            if bic <= bic_best:
                bic_best = bic
                k_best = k_denovo
                alpha_best = alpha
                beta_best = beta

        except Exception:
            continue
    
    return k_best, alpha_best, beta_best

'''




import numpy as np
import pandas as pd
import torch
import pyro
from pyro.infer import SVI,Trace_ELBO, JitTrace_ELBO, TraceEnum_ELBO
from pyro.ops.indexing import Vindex
from pyro.optim import Adam
import pyro.distributions.constraints as constraints
import pyro.distributions as dist
import torch.nn.functional as F
from tqdm import trange

import numpy as np
import pandas as pd
from statsmodels.tsa.stattools import adfuller
from statsmodels.tsa.stattools import kpss


class PyBasilica():

    def __init__(
        self,
        x,
        k_denovo,
        lr,
        n_steps,
        enumer=False, # if True, will enumerate the Z
        cluster=None,
        alpha_var=1,
        exp_rate=3,
        groups=None,
        beta_fixed=None,
        compile_model = True,
        CUDA = False,
        enforce_sparsity = False,
        store_parameters = False,
        regularizer = "cosine",
        reg_weight = 1,
        reg_bic = False, 
        stage = "random_noise", 
        regul_compare = None,
        seed = 10
        ):

        self._set_data_catalogue(x)
        self._set_beta_fixed(beta_fixed)
        self._set_k_denovo(k_denovo)

        self._set_hyperparams(enumer, cluster, alpha_var, exp_rate, groups)
        self._set_fit_settings(enforce_sparsity, lr, n_steps, compile_model, CUDA, regularizer, reg_weight, reg_bic, store_parameters, stage)

        self._fix_zero_denovo_null_reference()
        self._set_external_catalogue(regul_compare)

        self.seed = seed


    def _set_fit_settings(self, enforce_sparsity, lr, n_steps, compile_model, CUDA, \
                          regularizer, reg_weight, reg_bic, store_parameters, stage):
        self.enforce_sparsity = enforce_sparsity
        self.lr = lr
        self.n_steps = int(n_steps)
        self.compile_model = compile_model
        self.CUDA = CUDA
        self.regularizer = regularizer
        self.reg_weight = reg_weight
        self.reg_bic = reg_bic

        self.store_parameters = store_parameters
        self.stage = stage


    def _set_hyperparams(self, enumer, cluster, alpha_var, exp_rate, groups):
        self.enumer = enumer
        self.cluster = cluster
        self.alpha_var = alpha_var
        self.exp_rate = exp_rate

        self._set_groups(groups)

        self.init_params = None

        if not enumer and cluster != None:
            self.z_prior = torch.multinomial(torch.ones(cluster), self.n_samples, replacement=True).float()


    def _fix_zero_denovo_null_reference(self):
        if self.k_denovo == 0 and self.k_fixed == 0:
            self.stage = "random_noise"
            self.beta_fixed = torch.zeros(1, self.contexts, dtype=torch.float64)
            self.k_fixed = 1
            self._noise_only = True
        else:
            self._noise_only = False


    def _set_data_catalogue(self, x):
        try:
            self.x = torch.tensor(x.values, dtype=torch.float64)
            self.n_samples = x.shape[0]
            self.contexts = x.shape[1]
        except:
            raise Exception("Invalid mutations catalogue, expected Dataframe!")


    def _set_beta_fixed(self, beta_fixed):
        try:
            self.beta_fixed = torch.tensor(beta_fixed.values, dtype=torch.float64)
            if len(self.beta_fixed.shape)==1:
                self.beta_fixed = self.beta_fixed.reshape(1, self.beta_fixed.shape[0])

            self.k_fixed = beta_fixed.shape[0]
        except:
            if beta_fixed is None:
                self.beta_fixed = None
                self.k_fixed = 0
            else:
                raise Exception("Invalid fixed signatures catalogue, expected DataFrame!")
        
        if self.k_fixed > 0:
            self._fix_zero_contexts()


    def _fix_zero_contexts(self):
        colsums = torch.sum(self.beta_fixed, axis=0)
        zero_contexts = torch.where(colsums==0)[0]
        if torch.any(colsums == 0):
            # self.stage = "random_noise"
            random_sig = [0] if self.k_fixed == 1 else torch.randperm(self.beta_fixed.shape[0])[:torch.numel(zero_contexts)]

            for rr in random_sig:
                self.beta_fixed[rr, zero_contexts] = 1e-07

            self.beta_fixed = self.beta_fixed / (torch.sum(self.beta_fixed, 1).unsqueeze(-1))


    def _set_external_catalogue(self, regul_compare):
        try:
            self.regul_compare = torch.tensor(regul_compare.values, dtype=torch.float64)
        except:
            if regul_compare is None:
                self.regul_compare = None
            else:
                raise Exception("Invalid external signatures catalogue, expected DataFrame!")


        if self.k_fixed > 0:
            self._fix_zero_contexts()


    # def _fix_zero_contexts(self):
    #     colsums = torch.sum(self.beta_fixed, axis=0)
    #     zero_contexts = torch.where(colsums==0)[0]
    #     if torch.any(colsums == 0):
    #         # self.stage = "random_noise"
    #         random_sig = 0 if self.k_fixed == 1 else torch.randperm(self.beta_fixed.shape[0])[:torch.numel(zero_contexts)]

    #         for rr in random_sig:
    #             self.beta_fixed[rr, zero_contexts] = 1e-07

    #         self.beta_fixed = self.beta_fixed / (torch.sum(self.beta_fixed, 1).unsqueeze(-1))


    def _set_k_denovo(self, k_denovo):
        if isinstance(k_denovo, int):
            self.k_denovo = k_denovo
        else:
            raise Exception("Invalid k_denovo value, expected integer!")


    def _set_groups(self, groups):
        if groups is None:
            self.groups = None
        else:
            if isinstance(groups, list) and len(groups)==self.n_samples:
                self.groups = groups
            else:
                raise Exception("invalid groups argument, expected 'None' or a list with {} elements!".format(self.n_samples))


    def _check_args(self):
        pass
        # if self.k_denovo==0 and self.k_fixed==0:
        #     raise Exception("No. of denovo and fixed signatures could NOT be zero at the same time!")
        #if self.stage=="two":
        #    if self.alpha0 == None:
        #        raise Exception("stage two needs first stage exposure!")


    def model(self):

        n_samples = self.n_samples
        k_fixed = self.k_fixed
        k_denovo = self.k_denovo
        groups = self.groups
        cluster = self.cluster  # number of clusters or None
        enumer = self.enumer

        alpha_var = self.alpha_var
        exp_rate = self.exp_rate

        # Alpha
        if cluster != None:
            pi = pyro.sample("pi", dist.Dirichlet(torch.ones(cluster) / cluster))
            with pyro.plate("k1", k_fixed+k_denovo):
                with pyro.plate("g", cluster):
                    alpha_tissues = pyro.sample("alpha_t", dist.HalfNormal(alpha_var))

            with pyro.plate("n",n_samples):
                if enumer != False:
                    z = pyro.sample("latent_class", dist.Categorical(pi), infer={"enumerate":enumer})
                else:
                    z = pyro.sample("latent_class", dist.Categorical(pi)).long()
                alpha = pyro.sample("latent_exposure", dist.MultivariateNormal(alpha_tissues[z], torch.eye(k_fixed+k_denovo) * torch.tensor(alpha_var)))

        elif groups != None:

            n_groups = len(set(groups))
            if self._noise_only:
                alpha = torch.zeros(n_samples, 1, dtype=torch.float64)
            else:
                with pyro.plate("k1", k_fixed+k_denovo):
                    with pyro.plate("g", n_groups):
                        alpha_tissues = pyro.sample("alpha_t", dist.HalfNormal(alpha_var))

                # sample from the alpha prior
                with pyro.plate("k", k_fixed + k_denovo):   # columns
                    with pyro.plate("n", n_samples):        # rows
                        alpha = pyro.sample("latent_exposure", dist.Normal(alpha_tissues[groups,:], alpha_var))

        else:
            if self._noise_only:
                alpha = torch.zeros(n_samples, 1, dtype=torch.float64)
            else:
                with pyro.plate("k", k_fixed + k_denovo):   # columns
                    with pyro.plate("n", n_samples):        # rows
                        if self.enforce_sparsity:
                            alpha = pyro.sample("latent_exposure", dist.Exponential(exp_rate))
                        else:
                            alpha = pyro.sample("latent_exposure", dist.HalfNormal(alpha_var))

                alpha = alpha / (torch.sum(alpha, 1).unsqueeze(-1))
                alpha = torch.clamp(alpha, 0, 1)

        '''
        if self.stage=="two":
            dd = torch.tensor(self.alpha0.values).float()
            cc = torch.sum(dd, 1).unsqueeze(-1)
            alpha = alpha * ((torch.ones(alpha.shape[0]) - self.alpha0.sum) / self.alpha0.sum)
        '''

        # Epsilon
        if self.stage == "random_noise":
            eps_var = 10
            with pyro.plate("contexts3", self.contexts):  # columns
                    with pyro.plate("n3", n_samples):     # rows
                        epsilon = pyro.sample("latent_m", dist.HalfNormal(eps_var))
        else:
            epsilon = None

        # Beta
        if k_denovo == 0:
            beta_denovo = None
        else:
            with pyro.plate("contexts", self.contexts):  # columns
                with pyro.plate("k_denovo", k_denovo):  # rows
                    beta_denovo = pyro.sample("latent_signatures", dist.HalfNormal(1))
            
            beta_denovo = beta_denovo / (torch.sum(beta_denovo, 1).unsqueeze(-1))   # normalize
            beta_denovo = torch.clamp(beta_denovo, 0, 1)

        beta = self._get_unique_beta(self.beta_fixed, beta_denovo)
        reg = self._regularizer(self.beta_fixed, beta_denovo, self.regularizer)
        self.reg = reg

        # Observations
        with pyro.plate("contexts2", self.contexts):
            with pyro.plate("n2", n_samples):
                ## TODO might try to insert the alpha here

                a = torch.matmul(torch.matmul(torch.diag(torch.sum(self.x, axis=1)), alpha), beta)
                
                if self.stage == "random_noise":
                    xx = a + epsilon
                    lk =  dist.Poisson(xx).log_prob(self.x)
                else:
                    lk =  dist.Poisson(a).log_prob(self.x)
                # pyro.factor("loss", lk + self.reg_weight * (reg * self.x.shape[0] * self.x.shape[1]))
                pyro.factor("loss", lk.sum() + self.reg_weight * (reg * self.x.shape[0] * self.x.shape[1]))


    def _initialize_params(self):
        if self.init_params is None:
            params = dict()
            
            if self.cluster is not None:
                params["pi_param"] = torch.ones(self.cluster)
                params["alpha_t_param"] = dist.HalfNormal(torch.ones(self.cluster, self.k_fixed + self.k_denovo, dtype=torch.float6) * \
                                                        torch.tensor(self.alpha_var)).sample()
            elif self.groups is not None:
                params["alpha_t_param"] = dist.HalfNormal(torch.ones(len(set(self.groups)), self.k_fixed + self.k_denovo, dtype=torch.float64) * \
                                                        torch.tensor(self.alpha_var)).sample()
            else:
                if self.enforce_sparsity:
                    params["alpha_mean"] = dist.Exponential(torch.ones(self.n_samples, self.k_fixed + self.k_denovo, dtype=torch.float64) * self.exp_rate).sample()
                else:
                    params["alpha_mean"] = dist.HalfNormal(torch.ones(self.n_samples, self.k_fixed + self.k_denovo, dtype=torch.float64) * self.alpha_var).sample()

            params["epsilon_var"] = torch.ones(self.n_samples, self.contexts, dtype=torch.float64)

            if self.k_denovo > 0:
                params["beta_dn_param"] = dist.HalfNormal(torch.ones(self.k_denovo, self.contexts, dtype=torch.float64)).sample()

            self.init_params = params

        return self.init_params


    def guide(self):

        n_samples = self.n_samples
        k_fixed = self.k_fixed
        k_denovo = self.k_denovo
        groups = self.groups
        cluster = self.cluster
        enumer = self.enumer

        init_params = self._initialize_params()

        # Alpha
        if cluster is not None:
            pi_param = pyro.param("pi_param", init_params["pi_param"], constraint=constraints.simplex)
            pi = pyro.sample("pi", dist.Delta(pi_param).to_event(1))

            alpha_tissues = pyro.param("alpha_t_param", init_params["alpha_t_param"], constraint=constraints.greater_than_eq(0))

            with pyro.plate("k1", k_fixed+k_denovo):
                with pyro.plate("g", cluster):
                    pyro.sample("alpha_t", dist.Delta(alpha_tissues))

            if enumer == False:
                z_par = pyro.param("latent_class_p", lambda: self.z_prior)

            with pyro.plate("n", n_samples):  # + (n_samples) BATCH
                if enumer != False:
                    z = pyro.sample("latent_class", dist.Categorical(pi), infer={"enumerate":enumer})
                else:
                    z = pyro.sample("latent_class", dist.Delta(z_par)).long()
                    # Delta shape -> batch=(n_samples) + event=()
                    # adding plate dims -> BATCH : (k_denovo+k_fixed) (n_samples)
                alpha_p = pyro.param("alpha", lambda: alpha_tissues[z, :], constraint=constraints.greater_than_eq(0))
                alpha = pyro.sample("latent_exposure", dist.Delta(alpha_p).to_event(1))

        elif groups != None:
            if not self._noise_only:
                n_groups = len(set(groups))
                alpha_tissues = pyro.param("alpha_t_param", init_params["alpha_t_param"], constraint=constraints.greater_than_eq(0))

                with pyro.plate("k1", k_fixed+k_denovo):
                    with pyro.plate("g", n_groups):
                        pyro.sample("alpha_t", dist.Delta(alpha_tissues))

                with pyro.plate("k", k_fixed + k_denovo):   # columns
                    with pyro.plate("n", n_samples):        # rows
                        alpha = pyro.param("alpha", alpha_tissues[groups, :], constraint=constraints.greater_than_eq(0))
                        pyro.sample("latent_exposure", dist.Delta(alpha))

        # No groups
        else:
            if not self._noise_only:
                alpha_mean = init_params["alpha_mean"]

                with pyro.plate("k", k_fixed + k_denovo):
                    with pyro.plate("n", n_samples):
                        if self.enforce_sparsity:
                            alpha = pyro.param("alpha", alpha_mean, constraint=constraints.greater_than(0.0))
                        else:
                            alpha = pyro.param("alpha", alpha_mean, constraint=constraints.greater_than_eq(0.0))
                        pyro.sample("latent_exposure", dist.Delta(alpha))

        # Epsilon 
        if self.stage == "random_noise":
            eps_var = pyro.param("lambda_epsilon", init_params["epsilon_var"], constraint=constraints.positive)

            with pyro.plate("contexts3", self.contexts):
                with pyro.plate("n3", n_samples):
                    pyro.sample("latent_m", dist.HalfNormal(eps_var))

        # Beta 
        if k_denovo > 0:
            beta_par = init_params["beta_dn_param"]
            with pyro.plate("contexts", self.contexts):
                with pyro.plate("k_denovo", k_denovo):
                    beta = pyro.param("beta_denovo", beta_par, constraint=constraints.greater_than_eq(0.0))
                    pyro.sample("latent_signatures", dist.Delta(beta))


    def _regularizer(self, beta_fixed, beta_denovo, reg_type = "cosine"):
        '''
        if beta_denovo == None:
            dd = 0
        else:
            dd = 0
            c1 = 0
            for denovo1 in beta_denovo:
                c1 += 1
                c2 = 0
                for denovo2 in beta_denovo:
                    c2 += 1
                    if c1!=c2:
                        dd += F.kl_div(denovo1, denovo2, reduction="batchmean").item()
        '''
        loss = 0

        if self.regul_compare is not None:
            beta_fixed = self.regul_compare

        if beta_fixed is None or beta_denovo is None or self._noise_only:
            return loss
        
        beta_fixed[beta_fixed==0] = 1e-07

        if reg_type == "cosine":
            for fixed in beta_fixed:
                for denovo in beta_denovo:
                    loss += torch.log((1 - F.cosine_similarity(fixed, denovo, dim = -1)))
        elif reg_type == "KL":
            for fixed in beta_fixed:
                for denovo in beta_denovo:
                    loss += torch.log(F.kl_div(torch.log(fixed), torch.log(denovo), log_target = True, reduction="batchmean"))
        else:
            raise("The regularization admits either 'cosine' or 'KL'")

        return loss


    def _get_unique_beta(self, beta_fixed, beta_denovo):
        if beta_fixed is None: 
            return beta_denovo

        if beta_denovo is None or self._noise_only:
            return beta_fixed
        
        return torch.cat((beta_fixed, beta_denovo), axis=0)


    def _likelihood(self, M, alpha, beta_fixed, beta_denovo, eps_var=None):
        beta = self._get_unique_beta(beta_fixed, beta_denovo)

        a = torch.matmul(torch.matmul(torch.diag(torch.sum(M, axis=1)), alpha), beta)

        if eps_var == None:
            _log_like_matrix = dist.Poisson(a).log_prob(M)
        else:
            xx = a + dist.HalfNormal(eps_var).sample()
            _log_like_matrix = dist.Poisson(xx).log_prob(M)

        _log_like_sum = torch.sum(_log_like_matrix)
        _log_like = float("{:.3f}".format(_log_like_sum.item()))

        return _log_like


    def _initialize_seed(self, optim, elbo, seed):
        '''
        Auxiliary function to optimize the random initialization of the SVI object.
        '''
        pyro.set_rng_seed(seed)
        pyro.get_param_store().clear()

        svi = SVI(self.model, self.guide, optim, elbo)
        loss = svi.step()
        self.init_params = None

        return np.round(loss, 3), seed


    def _fit(self):
        pyro.clear_param_store()  # always clear the store before the inference
        if self.CUDA and torch.cuda.is_available():
            torch.set_default_tensor_type('torch.cuda.FloatTensor')
            self.x = self.x.cuda()
            if self.beta_fixed is not None:
                self.beta_fixed = self.beta_fixed.cuda()
        else:
            torch.set_default_tensor_type(t=torch.FloatTensor)

        if self.cluster != None and self.enumer != False:
            elbo = TraceEnum_ELBO()
        elif self.compile_model and not self.CUDA:
            elbo = JitTrace_ELBO()
        else:
            elbo = Trace_ELBO()

        min_steps = 50

        train_params = []
        # learning global parameters
        adam_params = {"lr": self.lr}
        optimizer = Adam(adam_params)

        _, self.seed = min([self._initialize_seed(optimizer, elbo, seed) for seed in range(50)], key = lambda x: x[0])
        pyro.set_rng_seed(self.seed)
        pyro.get_param_store().clear()

        svi = SVI(self.model, self.guide, optimizer, loss=elbo)

        losses = []
        regs = []
        likelihoods = []
        for _ in range(self.n_steps):   # inference - do gradient steps
            loss = svi.step()
            losses.append(loss)
            regs.append(self.reg)

            # create likelihoods 
            alpha = self._get_param("alpha", normalize=True, to_cpu=False)
            eps_var = self._get_param("eps_var", normalize=False, to_cpu=False)
            beta_denovo = self._get_param("beta_denovo", normalize=True, to_cpu=False)
            likelihoods.append(self._likelihood(self.x, alpha, self.beta_fixed, beta_denovo, eps_var))

            # convergence test 
            if len(losses) >= min_steps and len(losses) % min_steps == 0 and convergence(x=losses[-min_steps:], alpha=0.05):
                break

            if self.store_parameters:
                train_params.append(self.get_param_dict())

        if self.CUDA and torch.cuda.is_available():
          self.x = self.x.cpu()
          if self.beta_fixed is not None:
            self.beta_fixed = self.beta_fixed.cpu()

        self.train_params = train_params
        self.losses = losses
        self.likelihoods = likelihoods
        self.regs = regs
        self._set_params()
        self._set_bic()
        self.likelihood = self._likelihood(self.x, self.alpha, self.beta_fixed, self.beta_denovo, self.eps_var)


    def _set_params(self):
        self._set_alpha()
        self._set_beta_denovo()
        self._set_epsilon()
        self._set_clusters()


    def _get_param(self, param_name, normalize=False, to_cpu=True):
        try:
            par = pyro.param(param_name)

            if to_cpu and self.CUDA and torch.cuda.is_available(): par = par.cpu()

            try:
                par = par.clone().detach()
            finally:
                if normalize:
                    par = par / (torch.sum(par, 1).unsqueeze(-1))
        except:
            if self._noise_only and param_name=="alpha":
                # TODO check when running in GPU 
                return torch.zeros(self.n_samples, 1, dtype=torch.float64)
            return None

        return par


    def _set_alpha(self):
        self.alpha = self._get_param("alpha", normalize=True)
        self.alpha_unn = self._get_param("alpha", normalize=False)
        try:
            self.alpha_prior = self._get_param("alpha_t_param", normalize=True)
            self.alpha_prior_unn = self._get_param("alpha_t_param", normalize=False)
        except:
            self.alpha_prior = None
            self.alpha_prior_unn = None


    def _set_beta_denovo(self):
        if self.k_denovo == 0:
            self.beta_denovo = None
        else:
            self.beta_denovo = self._get_param("beta_denovo", normalize=True)


    def _set_epsilon(self):
        if self.stage=="random_noise":
            self.eps_var = self._get_param("lambda_epsilon", normalize=False)
        else:
            self.eps_var = None


    def _set_clusters(self):
        if self.cluster is None:
            return

        self.pi = self._get_param("pi", normalize=False)

        if self.enumer == False:
            self.z = pyro.param("latent_class_p")
            self.z = self._get_param("latent_class_p", normalize=False)
        else:
            self.z = self._compute_posterior_probs()


    def _logsumexp(self, weighted_lp) -> torch.Tensor:
        '''
        Returns `m + log( sum( exp( weighted_lp - m ) ) )`
        - `m` is the the maximum value of weighted_lp for each observation among the K values
        - `torch.exp(weighted_lp - m)` to perform some sort of normalization
        In this way the `exp` for the maximum value will be exp(0)=1, while for the
        others will be lower than 1, thus the sum across the K components will sum up to 1.
        '''
        m = torch.amax(weighted_lp, dim=0)  # the maximum value for each observation among the K values
        summed_lk = m + torch.log(torch.sum(torch.exp(weighted_lp - m), axis=0))
        return summed_lk


    def get_param_dict(self):
        params = dict()
        params["alpha"] = self._get_param("alpha", normalize=True)
        params["alpha_prior"] = self._get_param("alpha_t_param", normalize=True)

        params["beta_d"] =  self._get_param("beta_denovo", normalize=True)
        params["beta_f"] = self.beta_fixed

        params["pi"] = self._get_param("pi", normalize=False)

        if self.stage=="random_noise":
            params["lambda_epsilon"] = self._get_param("lambda_epsilon", normalize=False)
        else:
            params["lambda_epsilon"] = None

        return params


    def _compute_posterior_probs(self):
        params = self.get_param_dict()
        M = torch.tensor(self.x)
        cluster = self.cluster
        n_samples = self.n_samples

        try:
            beta = torch.cat((torch.tensor(params["beta_f"]), torch.tensor(params["beta_d"])), axis=0)
        except:
            beta = torch.tensor(params["beta_d"])

        z = torch.zeros(n_samples)

        for n in range(n_samples):
            m_n = M[n,:].unsqueeze(0)
            ll_nk = torch.zeros((cluster, M.shape[1]))

            for k in range(cluster):
                muts_n = torch.sum(m_n, axis=1).float()  # muts for patient n
                rate = torch.matmul( \
                    torch.matmul( torch.diag(muts_n), params["alpha_prior"][k,:].unsqueeze(0) ), \
                    beta.float() )

                # compute weighted log probability
                ll_nk[k,:] = torch.log(params["pi"][k]) + pyro.distributions.Poisson( rate ).log_prob(m_n)

            ll_nk_sum = ll_nk.sum(axis=1)  # sum over the contexts -> reaches a tensor of shape (n_clusters)

            ll = self._logsumexp(ll_nk_sum)
            probs = torch.exp(ll_nk_sum - ll)

            best_cl = torch.argmax(probs)
            z[n] = best_cl

        return z


    def _set_bic(self):

        M = self.x
        alpha = self.alpha

        _log_like = self._likelihood(M, alpha, self.beta_fixed, self.beta_denovo, self.eps_var)

        # adding regularizer
        if self.reg_bic:
            reg = self._regularizer(self.beta_fixed, self.beta_denovo, reg_type = self.regularizer)
            _log_like += self.reg_weight * (reg * self.x.shape[0] * self.x.shape[1])
        
        k = self._number_of_params()
        n = M.shape[0] * M.shape[1]
        bic = k * torch.log(torch.tensor(n, dtype=torch.float64)) - (2 * _log_like)

        self.bic = bic.item()

    
    def _number_of_params(self):
        if self.k_denovo == 0 and torch.sum(self.beta_fixed) == 0:
            k = 0
        else:
            k = (self.n_samples * (self.k_denovo + self.k_fixed)) + ((self.k_denovo) * self.contexts)
        
        if self.eps_var is not None:
            k = k + self.eps_var.shape[0] * self.eps_var.shape[1]
        
        return k


    def _convert_to_dataframe(self, x, beta_fixed):

        # mutations catalogue
        self.x = x
        sample_names = list(x.index)
        mutation_features = list(x.columns)

        # fixed signatures
        fixed_names = []
        if self.beta_fixed is not None and torch.sum(self.beta_fixed) > 0:
            fixed_names = list(beta_fixed.index)
            self.beta_fixed = beta_fixed

        # denovo signatures
        denovo_names = []
        if self.beta_denovo is not None:
            for d in range(self.k_denovo):
                denovo_names.append("D"+str(d+1))
            self.beta_denovo = pd.DataFrame(np.array(self.beta_denovo), index=denovo_names, columns=mutation_features)

        # alpha
        if len(fixed_names+denovo_names) > 0:
            self.alpha = pd.DataFrame(np.array(self.alpha), index=sample_names , columns=fixed_names + denovo_names)

        # epsilon variance
        if self.stage=="random_noise":
            self.eps_var = pd.DataFrame(np.array(self.eps_var), index=sample_names , columns=mutation_features)
        else:
            self.eps_var = None


    def _mv_to_gpu(self,*cpu_tens):
        [print(tens) for tens in cpu_tens]
        [tens.cuda() for tens in cpu_tens]


    def _mv_to_cpu(self,*gpu_tens):
        [tens.cpu() for tens in gpu_tens]





'''
Augmented Dicky-Fuller (ADF) test
* Null hypothesis (H0) — Time series is not stationary.
* Alternative hypothesis (H1) — Time series is stationary.

Kwiatkowski-Phillips-Schmidt-Shin test for stationarity
* Null hypothesis (H0) — Time series is stationary.
* Alternative hypothesis (H1) — Time series is not stationary.

both return tuples where 2nd value is P-value
'''


import warnings
warnings.filterwarnings('ignore')

def is_stationary(data: pd.Series, alpha: float = 0.05):

    # Test to see if the time series is already stationary
    if kpss(data, regression='c', nlags="auto")[1] > alpha:
    #if adfuller(data)[1] < alpha:
        # stationary - stop inference
        return True
    else:
        # non-stationary - continue inference
        return False

def convergence(x, alpha: float = 0.05):
    ### !!! REMEMBER TO CHECK !!! ###
    #return False
    if isinstance(x, list):
        data = pd.Series(x)
    else:
        raise Exception("input list is not valid type!, expected list.")

    return is_stationary(data, alpha=alpha)



    # def model(self):

    #     n_samples = self.n_samples
    #     k_fixed = self.k_fixed
    #     k_denovo = self.k_denovo
    #     groups = self.groups

    #     #----------------------------- [ALPHA] -------------------------------------
    #     if groups != None:

    #         #num_groups = max(params["groups"]) + 1
    #         n_groups = len(set(groups))
    #         alpha_tissues = dist.Normal(torch.zeros(n_groups, k_fixed + k_denovo), 1).sample()

    #         # sample from the alpha prior
    #         with pyro.plate("k", k_fixed + k_denovo):   # columns
    #             with pyro.plate("n", n_samples):        # rows
    #                 alpha = pyro.sample("latent_exposure", dist.Normal(alpha_tissues[groups, :], 1))
    #     else:
    #         alpha_mean = dist.Normal(torch.zeros(n_samples, k_fixed + k_denovo), 1).sample()

    #         with pyro.plate("k", k_fixed + k_denovo):   # columns
    #             with pyro.plate("n", n_samples):        # rows
    #                 if self.enforce_sparsity:
    #                     alpha = pyro.sample("latent_exposure", dist.Exponential(3))
    #                 else:
    #                     alpha = pyro.sample("latent_exposure", dist.HalfNormal(1))

    #     alpha = alpha / (torch.sum(alpha, 1).unsqueeze(-1))     # normalize
    #     alpha = torch.clamp(alpha, 0,1)

    #     #----------------------------- [BETA] -------------------------------------
    #     if k_denovo==0:
    #         beta_denovo = None
    #     else:
    #         #beta_mean = dist.Normal(torch.zeros(k_denovo, self.contexts), 1).sample()
    #         with pyro.plate("contexts", self.contexts):            # columns
    #             with pyro.plate("k_denovo", k_denovo):  # rows
    #                 beta_denovo = pyro.sample("latent_signatures", dist.HalfNormal(1))
    #         beta_denovo = beta_denovo / (torch.sum(beta_denovo, 1).unsqueeze(-1))   # normalize
    #         beta_denovo = torch.clamp(beta_denovo, 0,1)

    #     #----------------------------- [LIKELIHOOD] -------------------------------------
    #     if self.beta_fixed is None:
    #         beta = beta_denovo
    #         reg = 0
    #     elif beta_denovo is None:
    #         beta = self.beta_fixed
    #         reg = 0
    #     else:
    #         beta = torch.cat((self.beta_fixed, beta_denovo), axis=0)
    #         reg = self._regularizer(self.beta_fixed, beta_denovo)

    #     with pyro.plate("contexts2", self.contexts):
    #         with pyro.plate("n2", n_samples):
    #             lk =  dist.Poisson(torch.matmul(torch.matmul(torch.diag(torch.sum(self.x, axis=1)), alpha), beta)).log_prob(self.x)
    #             pyro.factor("loss", lk - reg)


    # def guide(self):

    #     n_samples = self.n_samples
    #     k_fixed = self.k_fixed
    #     k_denovo = self.k_denovo
    #     #groups = self.groups

    #     alpha_mean = dist.HalfNormal(torch.ones(n_samples, k_fixed + k_denovo)).sample()

    #     with pyro.plate("k", k_fixed + k_denovo):
    #         with pyro.plate("n", n_samples):
    #             alpha = pyro.param("alpha", alpha_mean, constraint=constraints.greater_than_eq(0))
    #             pyro.sample("latent_exposure", dist.Delta(alpha))

    #     if k_denovo != 0:
    #         beta_mean = dist.HalfNormal(torch.ones(k_denovo, self.contexts)).sample()
    #         with pyro.plate("contexts", self.contexts):
    #             with pyro.plate("k_denovo", k_denovo):
    #                 beta = pyro.param("beta_denovo", beta_mean, constraint=constraints.greater_than_eq(0))
    #                 pyro.sample("latent_signatures", dist.Delta(beta))



import random
import pyro.distributions as dist
import numpy as np
import pandas as pd
import torch
from statistics import mean


#-----------------------------------------------------------------[<QC-PASSED>]
# divide cosmic signatures into two parts:
# 1st part: considered as cosmic signatures
# 2nd part: considered as denovo signatures

def cosmic_denovo(full_cosmic_df):

    #random.seed(a=seed)

    # cosmic_path --- <class 'str'>
    #full_cosmic_df = pd.read_csv(cosmic_path, index_col=0)
    full_cosmic_list = list(full_cosmic_df.index)

    cosmic_list = random.sample(full_cosmic_list, k=50)
    denovo_list = list(set(full_cosmic_list) - set(cosmic_list))

    cosmic_df = full_cosmic_df.loc[cosmic_list] # <class 'pandas.core.frame.DataFrame'>
    denovo_df = full_cosmic_df.loc[denovo_list] # <class 'pandas.core.frame.DataFrame'>

    return cosmic_df, denovo_df
    # cosmic_df --- dtype: DataFrame
    # denovo_df --- dtype: DataFrame

#-----------------------------------------------------------------[QC:PASSED]

def target_generator(cosmic_df, denovo_df, target_complexity, num_samples):
    # profile ------- {"A", "B", "C"}
    # cosmic_df ----- <class 'pandas.core.frame.DataFrame'>
    # denovo_df ----- <class 'pandas.core.frame.DataFrame'>

    #random.seed(a=seed)

    # num_samples = random.randint(15, 25)
    num_samples = int(num_samples)

    # error handling for profile argument
    valid = ["low", "medium", "high"]
    if target_complexity not in valid:
        raise ValueError("profile must be one of %s." % valid)

    if target_complexity=="low":
        fixed_num = random.randint(3, 5)    # <class 'int'>
        denovo_num = random.randint(0, 2)
    elif target_complexity=="medium":
        fixed_num = random.randint(0, 2)
        denovo_num = random.randint(3, 5)
    elif target_complexity=="high":
        fixed_num = random.randint(3, 5)
        denovo_num = random.randint(3, 5)

    cosmic_list = list(cosmic_df.index)
    denovo_list = list(denovo_df.index)
    mutation_features = list(cosmic_df.columns)

    # beta fixed ------------------------------------------------------
    if fixed_num > 0:
        fixed_list = random.sample(cosmic_list, k=fixed_num)
        beta_fixed_df = cosmic_df.loc[fixed_list]
    else:
        beta_fixed_df = pd.DataFrame(columns=mutation_features)
        #beta_fixed_df = None
    
    # beta denovo -----------------------------------------------------
    if denovo_num > 0:
        denovo_list = random.sample(denovo_list, k=denovo_num)
        beta_denovo_df = denovo_df.loc[denovo_list]

        # add "_D" to the end of denovo signatures to distinguish them from fixed signatures
        denovo_labels = []
        for i in range(len(denovo_list)):
            denovo_labels.append(denovo_list[i] + '_D')
        beta_denovo_df.index = denovo_labels
    else:
        beta_denovo_df = pd.DataFrame(columns=mutation_features)
        #beta_denovo_df = None

    if beta_denovo_df.empty:
        beta_df = beta_fixed_df
    elif beta_fixed_df.empty:
        beta_df = beta_denovo_df
    else:
        beta_df = pd.concat([beta_fixed_df, beta_denovo_df], axis=0)

    signatures = list(beta_df.index)
    beta_tensor = torch.tensor(beta_df.values).float()

    #------- alpha ----------------------------------------------
    matrix = np.random.rand(num_samples, len(signatures))
    alpha_tensor = torch.tensor(matrix / matrix.sum(axis=1)[:, None]).float()
    alpha_np = np.array(alpha_tensor)
    alpha_df = pd.DataFrame(alpha_np, columns=signatures)

    #------- theta ----------------------------------------------
    theta = random.sample(range(1000, 4000), k=num_samples) # dtype:list
    
    #------- check dimensions -----------------------------------
    m_alpha = alpha_tensor.size()[0]    # no. of branches (alpha)
    m_theta = len(theta)                # no. of branches (theta)
    k_alpha = alpha_tensor.size()[1]    # no. of signatures (alpha)
    k_beta = beta_tensor.size()[0]      # no. of signatures (beta)
    if not(m_alpha == m_theta and k_alpha == k_beta):
        raise ValueError("wrong dimensions!")
    
    M_tensor = torch.zeros([num_samples, 96])   # initialize mutational catalogue with zeros

    for i in range(num_samples):
        p = alpha_tensor[i]         # selecting branch i
        for k in range(theta[i]):   # iterate for number of the mutations in branch i

            # sample signature profile index from categorical data
            b = beta_tensor[dist.Categorical(p).sample().item()]

            # sample mutation feature index for corresponding signature from categorical data
            j = dist.Categorical(b).sample().item()

            # add +1 to the mutation feature in position j in branch i
            M_tensor[i, j] += 1

    # phylogeny
    M_np = np.array(M_tensor)
    M_df = pd.DataFrame(M_np, columns=mutation_features)
    M_df = M_df.astype(int)

    # return all in dataframe format
    return M_df, alpha_df, beta_fixed_df, beta_denovo_df
    # M_df ------------- dtype: dataframe
    # alpha_df --------- dtype: dataframe
    # beta_fixed_df ---- dtype: dataframe, could be empty (if empty --> beta_denovo_df != empty)
    # beta_denovo_df --- dtype: dataframe, could be empty (if empty --> beta_fixed_df != empty)

#-----------------------------------------------------------------[QC:PASSED]
def input_catalogue_generator(cosmic_df, beta_fixed_df, beta_denovo_df, input_complexity):
    # profile ---------- {"X", "Y", "Z"}
    # beta_fixed_df ---- dtype: dataframe
    # beta_denovo_df --- dtype: dataframe
    # cosmic_df -------- dtype: dataframe

    #random.seed(a=seed)

    # error handling for profile argument
    valid = ["low", "medium", "high"]
    if input_complexity not in valid:
        raise ValueError("profile must be one of %s." % valid)

    # TARGET DATA -----------------------------------------------------------------------------
    beta_fixed_names = list(beta_fixed_df.index)    # target beta signatures names (dtype: list)
    k_fixed_target = len(beta_fixed_names)
    beta_denovo_names = list(beta_denovo_df.index)  # target beta signatures names (dtype: list)
    k_denovo_target = len(beta_denovo_names)

    # common fixed signatures (target intersect test)
    # different fixed signatures (test minus target)
    if input_complexity=="low":
        if k_fixed_target > 0:
            k_overlap = random.randint(1, k_fixed_target)
            k_extra = 0
        else:
            k_overlap = 0
            k_extra = random.randint(1, k_denovo_target)

    elif input_complexity=="medium":
        if k_fixed_target > 0:
            k_overlap = random.randint(1, k_fixed_target)
            k_extra = random.randint(1, k_fixed_target)
        else:
            k_overlap = 0
            k_extra = random.randint(1, k_denovo_target)

    elif input_complexity=="high":
        k_overlap = 0
        k_extra = random.randint(1, k_fixed_target + k_denovo_target)
    
    
    cosmic_names = list(cosmic_df.index)    # cosmic signatures names (dtype: list)

    # exclude beta target fixed signatures from cosmic
    for signature in beta_fixed_names:
        cosmic_names.remove(signature)
    
    # common fixed signatures list
    if k_overlap > 0:
        overlap_sigs = random.sample(beta_fixed_names, k=k_overlap)
    else:
        overlap_sigs = []
    
    # different fixed signatures list
    if k_extra > 0:
        extra_sigs = random.sample(cosmic_names, k=k_extra)
    else:
        extra_sigs = []
    
    beta_input = cosmic_df.loc[overlap_sigs + extra_sigs]

    return beta_input   # dtype: dataframe


#-----------------------------------------------------------------[PASSED]
def input_generator(full_cosmic_df, target_complexity, input_complexity, num_samples):

    cosmic_df, denovo_df = cosmic_denovo(full_cosmic_df)
    
    M_df, alpha_df, beta_fixed_df, beta_denovo_df = target_generator(cosmic_df, denovo_df, target_complexity, num_samples)
    
    beta_input_df = input_catalogue_generator(cosmic_df, beta_fixed_df, beta_denovo_df, input_complexity)
    
    #beta_df = pd.concat([beta_fixed_df, beta_denovo_df], axis=0)

    if beta_fixed_df.empty:
        beta_fixed_df = None

    if beta_denovo_df.empty:
        beta_denovo_df = None

    if beta_input_df.empty:
        beta_input_df = None

    data = {
        "M" : M_df,                         # dataframe
        "alpha" : alpha_df,                 # dataframe
        "beta_fixed" : beta_fixed_df,       # dataframe
        "beta_denovo" : beta_denovo_df,     # dataframe
        "beta_input" : beta_input_df,       # dataframe
        "cosmic_df" : cosmic_df             # dataframe
    }

    return data





from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name = "pybasilica",
    version = "0.1.4",
    author = "Azad Sadr",
    author_email = "azad.sadr.h@gmail.com",
    description = "A bayesian model to extract mutational signatures",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = "https://github.com/caravagnalab/pybasilica",
    packages = ["pybasilica"],
    python_requires = ">=3.8",
    install_requires = [
        "pandas>=1.4.2",
        "pyro-api>=0.1.2",
        "pyro-ppl>=1.8.0",
        "numpy>=1.21.5",
        "torch==1.10.2",
        "tqdm",
        "rich",
        "statsmodels",
        "uniplot"
        ],
)
